import java.util.ArrayList;
import java.util.Iterator;

public class StateDiagram {

  private ArrayList<State> states;
  private ArrayList<Transition> transitions;
  private boolean error;

  private static StateDiagram theInstance = null;

  private StateDiagram() {
    states = new ArrayList<State>();
    transitions = new ArrayList<Transition>();

  }

  public static StateDiagram getIntance() {
    if (theInstance == null)
      theInstance = new StateDiagram();
    return theInstance;
  }
  
  public boolean getError(){
    return this.error;
  }

  //deletes the current state diagram
  public void dump() {
    states.clear();
    transitions.clear();
  }

  //returns a state search by its internal id
  public State getStatebyId(int id) {
    Iterator<State> i = states.iterator();
    while (i.hasNext()) {
      State s = i.next();
      if (s.getId() == id)
        return s;
    }
    return null;
  }

  //returns a transition searched by its internal id
  public Transition getTransitionbyId(int id) {
    Iterator<Transition> i = transitions.iterator();
    while (i.hasNext()) {
      Transition t = i.next();
      if (t.getId() == id)
        return t;
    }
    return null;
  }

  public ArrayList<State> getStates() {
    return states;
  }

  public ArrayList<Transition> getTransitions() {
    return transitions;
  }

  public String toString() {
    String s = "States:\n";
    Iterator<State> is = states.iterator();
    while (is.hasNext()) {
      State st = is.next();
      s += "Name: " + st.getName() + "\n";
      s += "Id: " + st.getCodeid() + "\n";
      s += "outgoing transitions: ";
      Iterator<Integer> ii = st.getTransitions_out().iterator();
      while (ii.hasNext()) {
        int out = ii.next();
        s += out + "; ";
      }
      s += "\n";
      s += "---------------\n";
    }
    s += "Transitions:\n";
    Iterator<Transition> it = transitions.iterator();
    while (it.hasNext()) {
      Transition t = it.next();
      s += "Name: " + t.getName() + "\n";
      s += "Id: " + t.getId() + "\n";
      s += "Start: " + t.getStart().getName() + " ("
          + t.getStart().getCodeid() + ")\n";
      s += "End: " + t.getEnd().getName() + " (" + t.getEnd().getCodeid()
          + ")\n";
      s += "Trigger: " + t.getTrigger() + "\n";
      s += "Guard: " + t.getGuard() + "\n";
      s += "Effect: " + t.getEffect() + "\n";
      s += "..............\n";
    }
    return s;
  }

  //returns the initial state. if there exists more than one the first one is returned 
  private int getInitialState() {
    Iterator<State> is = getStates().iterator();
    while (is.hasNext()) {
      State s = is.next();
      if (s.getName().contains("Initial")) {
        return s.getCodeid();
      }
    }
    return -1;
  }

  //generate the arduino-code
  public String generateCode(String[] pins) {
    error= false;
    String code = "";
    // kommentar zu nummern
    code += "/*\n";
    code += "Zustnde und Nummern:\n";
    Iterator<State> is = getStates().iterator();
    while (is.hasNext()) {
      State s = is.next();
      code += s.getCodeid() + ": " + s.getName() + "\n";
    }
    code += "*/\n";
    code += "int Zustandsnummer;\n";
    code += "int timer;\n";
    code += "\n";
    code += setupMethod(pins);
    code += "\n";
    code += loopMethod();

    return code;
  }

  //generates the setup method of the arduino code
  private String setupMethod(String[] pins) {
    String res = "";
    res += "void setup(){\n";
    for (int i = 0; i < pins.length; i++) {
      if (!pins[i].equals("unbenutzt")) {
        res += "pinMode(" + (i+1) + ",";
        if (pins[i].equals("Eingang"))
          res += "INPUT);\n";
        else
          res += "OUTPUT);\n";
      }
    }
    res += "timer=0;\n";
    res += "Zustandsnummer=" + getInitialState() + ";\n";

    res += "}\n";
    return res;

  }

//generates the loop method of the arduino code
  private String loopMethod() {
    String loop = "";

    loop += "void loop(){\n";
    loop += "switch(Zustandsnummer) {\n";

    Iterator<State> is = states.iterator();
    while (is.hasNext()) {
      State st = is.next();
      loop += "  case " + st.getCodeid() + ": {";
      switch (checkcomplexity(st.getTransitions_out())) {
      case -2:
        loop += "break;}\n";
        break;
      case -1:
        loop += complextransition(st.getTransitions_out());
        loop += "}\n";
        break;
      case 0:
        loop += singletransition(st.getTransitions_out());
        loop += "}\n";
        break;
      case 1:
        loop += multipletransition2(st.getTransitions_out());
        loop += "}\n";
        break;
      case 2:
        loop += multipletransition1(st.getTransitions_out());
        loop += "}\n";
        break;
      }
    }

    loop += "}\n";
    loop += "}";
    return loop;
  }

  //checks the complexity of a state considering the quantity of outgoing transitions  
  private int checkcomplexity(ArrayList<Integer> transitions) {
    if (transitions.size() == 0) { //something like end-transition
      return -2;
    }
    if (transitions.size() == 1) { //only one outgoing transition
      return 0;
    }
    if (transitions.size() > 1) { //if there exists more than one outgoing transition the complexity is measured by considering the quantity of found triggers
      int triggercount = 0;
      Iterator<Integer> ii = transitions.iterator();
      while (ii.hasNext()) {
        Transition t = getTransitionbyId(ii.next());
        if (t.getTrigger() != null)
          triggercount += 1;
      }
      if (triggercount == 1) {
        return 1;
      }
      if (triggercount == 0) {
        return 2; 
      }
    }
    return -1; //no considered case found. transition can't be clearly recognised
  }


  private String complextransition(ArrayList<Integer> transitions) {
    String res = "Die Bedeutung der bergnge konnte nicht erkannt werden";
    error=true;
    return res;
  }


  private String singletransition(ArrayList<Integer> transitions) {
    String res = "";
    if (transitions.size() == 1) {
      Transition t = getTransitionbyId(transitions.get(0));
      if (t.getGuard() == null && t.getTrigger() == null
          && t.getEffect() == null) {
        res = "Zustandsnummer=" + t.getEnd().getCodeid() + ";"+"break;";
      }
      if (t.getGuard() == null && t.getTrigger() == null
          && t.getEffect() != null) {
        res = t.getEffect();
        res += " Zustandsnummer=" + t.getEnd().getCodeid() + ";"+"break;";
      }
      if (t.getGuard() != null && t.getTrigger() == null) {
        res = "if (" + t.getGuard() + ")";
        if (t.getEffect() != null)
          res += "{" + t.getEffect() + "} ";
        res += "Zustandsnummer=" + t.getEnd().getCodeid() + ";"+"break;";
      }
      if (t.getTrigger() != null && t.getGuard() == null) {
        res = "delay(" + extractmillis(t.getTrigger()) + "); ";
        if (t.getEffect() != null)
          res += t.getEffect() + " ";
        res += "Zustandsnummer=" + t.getEnd().getCodeid() + ";"+"break;";
      }
    }
    return res;
  }

  private String multipletransition1(ArrayList<Integer> transitions) {
    String res = "";
    Transition t = null;
    Iterator<Integer> ii = transitions.iterator();
    while (ii.hasNext()) {
      t = StateDiagram.getIntance().getTransitionbyId(ii.next());
      res += "if (" + t.getGuard() + ") {";
      if (t.getEffect()!=null) res+=t.getEffect();
      res += " Zustandsnummer=" + t.getEnd().getCodeid() + "; break;} ";
    }
    return res;
  }

  private String multipletransition2(ArrayList<Integer> transitions) {
    String res = "";
    String ifblock = "";
    Transition t = null;
    Transition triggert = null;
    int triggertime;
    Iterator<Integer> ii = transitions.iterator();
    while (ii.hasNext()) {
      t = StateDiagram.getIntance().getTransitionbyId(ii.next());
      if (t.getTrigger() != null) {
        triggert = t;
      } else {
        ifblock += "if (" + t.getGuard() + ") {";
        if (t.getEffect()!=null) ifblock+= t.getEffect();
        ifblock+= " Zustandsnummer=" + t.getEnd().getCodeid() + "; break;}\n";
      }

    }
    triggertime = Integer.parseInt(extractmillis(triggert.getTrigger())) / 10;
    res += "Zustandsnummer=" + triggert.getEnd().getCodeid() + ";\n";
    res += "timer=0;\n";
    res += "while(timer!=" + triggertime + "){\n";
    res += "timer=timer+1;\ndelay(10);\n";
    res += ifblock;
    res += "}\n";
    res += "if (Zustandsnummer==" + triggert.getEnd().getCodeid() + ") {";
    if (triggert.getEffect()!=null) res+=triggert.getEffect();
    res+=" break;}\n";
    return res;
  }

  // transformes the format: after xx:yy:zz into milliseconds
  private String extractmillis(String trigger) {
    int hour = 0, min = 0, sec = 0, millis;
    if (trigger.startsWith("after ")) {
      hour = Integer.parseInt(trigger.substring(
          trigger.indexOf("after ") + 6,
          trigger.indexOf("after ") + 8));
      min = Integer.parseInt(trigger.substring(trigger.indexOf(":") + 1,
          trigger.indexOf(":") + 3));
      sec = Integer
          .parseInt(trigger.substring(trigger.lastIndexOf(":") + 1,
              trigger.lastIndexOf(":") + 3));
      // System.out.println(hour+";"+min+";"+sec);
    }
    millis = (sec + min * 60 + hour * 3600) * 1000;
    return Integer.toString(millis);
  }

}